"""Tests for motion sensors."""

import collections
import pathlib
import sys
import unittest
import unittest.mock

import HABApp

import habapp_rules.core.exceptions
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.motion
import habapp_rules.sensors.motion
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base


class TestMotion(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing motion sensors rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion_min_raw", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion_min_filtered", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion_max_raw", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion_max_filtered", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Motion_max_lock", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Sleep_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Brightness", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Brightness_Threshold", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Motion_Unittest_Motion_min_raw_state", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "CustomState", None)

        config_min = habapp_rules.sensors.config.motion.MotionConfig(
            items=habapp_rules.sensors.config.motion.MotionItems(motion_raw="Unittest_Motion_min_raw", motion_filtered="Unittest_Motion_min_filtered", state="H_Motion_Unittest_Motion_min_raw_state")
        )

        config_max = habapp_rules.sensors.config.motion.MotionConfig(
            items=habapp_rules.sensors.config.motion.MotionItems(
                motion_raw="Unittest_Motion_max_raw",
                motion_filtered="Unittest_Motion_max_filtered",
                brightness="Unittest_Brightness",
                brightness_threshold="Unittest_Brightness_Threshold",
                sleep_state="Unittest_Sleep_state",
                lock="Unittest_Motion_max_lock",
                state="CustomState",
            ),
            parameter=habapp_rules.sensors.config.motion.MotionParameter(extended_motion_time=5),
        )

        self.motion_min = habapp_rules.sensors.motion.Motion(config_min)
        self.motion_max = habapp_rules.sensors.motion.Motion(config_max)

    def test__init__(self) -> None:
        """Test __init__."""
        expected_states = [
            {"name": "Locked"},
            {"name": "SleepLocked"},
            {"name": "PostSleepLocked", "timeout": 99, "on_timeout": "timeout_post_sleep_locked"},
            {"name": "Unlocked", "initial": "Init", "children": [{"name": "Init"}, {"name": "Wait"}, {"name": "Motion"}, {"name": "MotionExtended", "timeout": 99, "on_timeout": "timeout_motion_extended"}, {"name": "TooBright"}]},
        ]
        self.assertEqual(expected_states, self.motion_min.states)

        expected_trans = [
            {"trigger": "lock_on", "source": ["Unlocked", "SleepLocked", "PostSleepLocked"], "dest": "Locked"},
            {"trigger": "lock_off", "source": "Locked", "dest": "Unlocked", "unless": "_sleep_active"},
            {"trigger": "lock_off", "source": "Locked", "dest": "SleepLocked", "conditions": "_sleep_active"},
            {"trigger": "sleep_started", "source": ["Unlocked", "PostSleepLocked"], "dest": "SleepLocked"},
            {"trigger": "sleep_end", "source": "SleepLocked", "dest": "Unlocked", "unless": "_post_sleep_lock_configured"},
            {"trigger": "sleep_end", "source": "SleepLocked", "dest": "PostSleepLocked", "conditions": "_post_sleep_lock_configured"},
            {"trigger": "timeout_post_sleep_locked", "source": "PostSleepLocked", "dest": "Unlocked", "unless": "_raw_motion_active"},
            {"trigger": "motion_off", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
            {"trigger": "motion_on", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
            {"trigger": "motion_on", "source": "Unlocked_Wait", "dest": "Unlocked_Motion"},
            {"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_MotionExtended", "conditions": "_motion_extended_configured"},
            {"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_Wait", "unless": "_motion_extended_configured"},
            {"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Wait", "unless": "_brightness_over_threshold"},
            {"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_TooBright", "conditions": "_brightness_over_threshold"},
            {"trigger": "motion_on", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Motion"},
            {"trigger": "brightness_over_threshold", "source": "Unlocked_Wait", "dest": "Unlocked_TooBright"},
            {"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Wait", "unless": "_raw_motion_active"},
            {"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Motion", "conditions": "_raw_motion_active"},
        ]
        self.assertEqual(expected_trans, self.motion_min.trans)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Motion"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        motion_graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.motion_min.states, transitions=self.motion_min.trans, initial=self.motion_min.state, show_conditions=True)

        motion_graph.get_graph().draw(picture_dir / "Motion.png", format="png", prog="dot")

    def test_initial_state(self) -> None:
        """Test _get_initial_state."""
        tests.helper.oh_item.item_state_change_event("Unittest_Brightness", 100)
        tests.helper.oh_item.item_state_change_event("Unittest_Brightness_Threshold", 1000)

        TestCase = collections.namedtuple("TestCase", "locked, sleep_state, brightness, motion_raw, expected_state_max, expected_state_min")

        test_cases = [
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=500, motion_raw=False, expected_state_max="Unlocked_Wait", expected_state_min="Unlocked_Wait"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=500, motion_raw=True, expected_state_max="Unlocked_Motion", expected_state_min="Unlocked_Motion"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=1500, motion_raw=False, expected_state_max="Unlocked_TooBright", expected_state_min="Unlocked_Wait"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=1500, motion_raw=True, expected_state_max="Unlocked_TooBright", expected_state_min="Unlocked_Motion"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=500, motion_raw=False, expected_state_max="SleepLocked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=500, motion_raw=True, expected_state_max="SleepLocked", expected_state_min="Unlocked_Motion"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=1500, motion_raw=False, expected_state_max="SleepLocked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=False, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=1500, motion_raw=True, expected_state_max="SleepLocked", expected_state_min="Unlocked_Motion"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=500, motion_raw=False, expected_state_max="Locked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=500, motion_raw=True, expected_state_max="Locked", expected_state_min="Unlocked_Motion"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=1500, motion_raw=False, expected_state_max="Locked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.AWAKE.value, brightness=1500, motion_raw=True, expected_state_max="Locked", expected_state_min="Unlocked_Motion"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=500, motion_raw=False, expected_state_max="Locked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=500, motion_raw=True, expected_state_max="Locked", expected_state_min="Unlocked_Motion"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=1500, motion_raw=False, expected_state_max="Locked", expected_state_min="Unlocked_Wait"),
            TestCase(locked=True, sleep_state=habapp_rules.system.SleepState.SLEEPING.value, brightness=1500, motion_raw=True, expected_state_max="Locked", expected_state_min="Unlocked_Motion"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Motion_max_lock", "ON" if test_case.locked else "OFF")
                tests.helper.oh_item.set_state("Unittest_Sleep_state", test_case.sleep_state)
                tests.helper.oh_item.set_state("Unittest_Brightness", test_case.brightness)
                tests.helper.oh_item.set_state("Unittest_Motion_max_raw", "ON" if test_case.motion_raw else "OFF")
                tests.helper.oh_item.set_state("Unittest_Motion_min_raw", "ON" if test_case.motion_raw else "OFF")

                self.assertEqual(test_case.expected_state_max, self.motion_max._get_initial_state("test"))
                self.assertEqual(test_case.expected_state_min, self.motion_min._get_initial_state("test"))

    def test_raw_motion_active(self) -> None:
        """Test _raw_motion_active."""
        tests.helper.oh_item.set_state("Unittest_Motion_min_raw", "ON")
        self.assertTrue(self.motion_min._raw_motion_active())

        tests.helper.oh_item.set_state("Unittest_Motion_min_raw", "OFF")
        self.assertFalse(self.motion_min._raw_motion_active())

    def test_get_brightness_threshold(self) -> None:
        """Test _get_brightness_threshold."""
        # value of threshold item
        self.assertEqual(float("inf"), self.motion_max._get_brightness_threshold())

        # value given as parameter
        self.motion_max._config.parameter.brightness_threshold = 800
        self.assertEqual(800, self.motion_max._get_brightness_threshold())

    def test_get_brightness_threshold_exceptions(self) -> None:
        """Test exceptions of _get_brightness_threshold."""
        self.motion_max._config.items.brightness_threshold = None
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesError):
            self.motion_max._get_brightness_threshold()

    def test_initial_unlock_state(self) -> None:
        """Test initial state of unlock state."""
        self.assertEqual(float("inf"), self.motion_max._get_brightness_threshold())
        tests.helper.oh_item.item_state_change_event("Unittest_Brightness", 100)
        tests.helper.oh_item.item_state_change_event("Unittest_Brightness_Threshold", 1000)
        self.assertEqual(1000, self.motion_max._get_brightness_threshold())

        TestCase = collections.namedtuple("TestCase", "brightness_value, motion_raw, expected_state_min, expected_state_max")

        test_cases = [
            TestCase(100, False, expected_state_min="Unlocked_Wait", expected_state_max="Unlocked_Wait"),
            TestCase(100, True, expected_state_min="Unlocked_Motion", expected_state_max="Unlocked_Motion"),
            TestCase(2000, False, expected_state_min="Unlocked_Wait", expected_state_max="Unlocked_TooBright"),
            TestCase(2000, True, expected_state_min="Unlocked_Motion", expected_state_max="Unlocked_TooBright"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Brightness", test_case.brightness_value)
                tests.helper.oh_item.set_state("Unittest_Motion_min_raw", "ON" if test_case.motion_raw else "OFF")
                tests.helper.oh_item.set_state("Unittest_Motion_max_raw", "ON" if test_case.motion_raw else "OFF")

                self.motion_min.to_Unlocked()
                self.motion_max.to_Unlocked()

                self.assertEqual(test_case.expected_state_min, self.motion_min.state)
                self.assertEqual(test_case.expected_state_max, self.motion_max.state)

    def test_lock(self) -> None:
        """Test if lock is activated from all states."""
        for state in self._get_state_names(self.motion_max.states):
            tests.helper.oh_item.set_state("Unittest_Motion_max_lock", "OFF")
            self.motion_max.state = state
            tests.helper.oh_item.send_command("Unittest_Motion_max_lock", "ON", "OFF")
            self.assertEqual("Locked", self.motion_max.state)

    def test_motion_extended_configured(self) -> None:
        """Test _motion_extended_configured."""
        self.motion_max._config.parameter.extended_motion_time = -1
        self.assertFalse(self.motion_max._motion_extended_configured())

        self.motion_max._config.parameter.extended_motion_time = 0
        self.assertFalse(self.motion_max._motion_extended_configured())

        self.motion_max._config.parameter.extended_motion_time = 1
        self.assertTrue(self.motion_max._motion_extended_configured())

    def test_post_sleep_lock_configured(self) -> None:
        """Test _post_sleep_lock_configured."""
        self.motion_max._config.parameter.post_sleep_lock_time = -1
        self.assertFalse(self.motion_max._post_sleep_lock_configured())

        self.motion_max._config.parameter.post_sleep_lock_time = 0
        self.assertFalse(self.motion_max._post_sleep_lock_configured())

        self.motion_max._config.parameter.post_sleep_lock_time = 1
        self.assertTrue(self.motion_max._post_sleep_lock_configured())

    def test_sleep_active(self) -> None:
        """Test _sleep_active."""
        tests.helper.oh_item.set_state("Unittest_Sleep_state", habapp_rules.system.SleepState.AWAKE.value)
        self.assertFalse(self.motion_max._sleep_active())

        tests.helper.oh_item.set_state("Unittest_Sleep_state", habapp_rules.system.SleepState.SLEEPING.value)
        self.assertTrue(self.motion_max._sleep_active())

    def test_transitions_locked(self) -> None:
        """Test leaving transitions of locked state."""
        # to Unlocked
        self.motion_max.state = "Locked"
        with unittest.mock.patch.object(self.motion_max, "_sleep_active", return_value=False):
            tests.helper.oh_item.send_command("Unittest_Motion_max_lock", "OFF", "ON")
        self.assertEqual("Unlocked_Wait", self.motion_max.state)

        # to SleepLocked
        self.motion_max.state = "Locked"
        with unittest.mock.patch.object(self.motion_max, "_sleep_active", return_value=True):
            tests.helper.oh_item.send_command("Unittest_Motion_max_lock", "OFF", "ON")
        self.assertEqual("SleepLocked", self.motion_max.state)

    def test_transitions_sleep_locked(self) -> None:
        """Test leaving transitions of sleep locked state."""
        # to Unlocked
        self.motion_max.state = "SleepLocked"
        with unittest.mock.patch.object(self.motion_max, "_post_sleep_lock_configured", return_value=False):
            self.motion_max.sleep_end()
        self.assertEqual("Unlocked_Wait", self.motion_max.state)

        # to PostSleepLocked
        self.motion_max.state = "SleepLocked"
        with unittest.mock.patch.object(self.motion_max, "_post_sleep_lock_configured", return_value=True):
            self.motion_max.sleep_end()
        self.assertEqual("PostSleepLocked", self.motion_max.state)

    def test_transitions_post_sleep_locked(self) -> None:
        """Test leaving transitions of post sleep locked state."""
        # to Unlocked | motion not active
        self.motion_max.state = "PostSleepLocked"
        with unittest.mock.patch.object(self.motion_max, "_raw_motion_active", return_value=False):
            self.motion_max.timeout_post_sleep_locked()
            self.assertEqual("Unlocked_Wait", self.motion_max.state)

        # no change after timeout and motion
        self.motion_max.state = "PostSleepLocked"
        with unittest.mock.patch.object(self.motion_max, "_raw_motion_active", return_value=True):
            self.motion_max.timeout_post_sleep_locked()
            self.assertEqual("PostSleepLocked", self.motion_max.state)

        # reset timer if motion off
        self.motion_max.state = "PostSleepLocked"
        self.transitions_timer_mock.reset_mock()
        tests.helper.oh_item.item_state_change_event("Unittest_Motion_max_raw", "OFF", "ON")
        self.assertEqual("PostSleepLocked", self.motion_max.state)
        self.assertEqual(1, self.transitions_timer_mock.call_count)

        # reset timer if motion on
        self.motion_max.state = "PostSleepLocked"
        self.transitions_timer_mock.reset_mock()
        tests.helper.oh_item.item_state_change_event("Unittest_Motion_max_raw", "ON", "OFF")
        self.assertEqual("PostSleepLocked", self.motion_max.state)
        self.assertEqual(1, self.transitions_timer_mock.call_count)

        # sleep starts during post sleep
        self.motion_max.state = "PostSleepLocked"
        tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", habapp_rules.system.SleepState.SLEEPING.value)
        self.assertEqual("SleepLocked", self.motion_max.state)

    def test_unlocked_wait(self) -> None:
        """Test leaving transitions of Unlocked_Wait state."""
        # to motion
        self.motion_max.state = "Unlocked_Wait"
        self.motion_max.motion_on()
        self.assertEqual("Unlocked_Motion", self.motion_max.state)

        # to TooBright
        self.motion_max.state = "Unlocked_Wait"
        self.motion_max.brightness_over_threshold()
        self.assertEqual("Unlocked_TooBright", self.motion_max.state)

    def test_unlocked_motion(self) -> None:
        """Test leaving transitions of Unlocked_Motion state."""
        # motion off | extended active
        self.motion_max.state = "Unlocked_Motion"
        with unittest.mock.patch.object(self.motion_max, "_motion_extended_configured", return_value=True):
            self.motion_max.motion_off()
        self.assertEqual("Unlocked_MotionExtended", self.motion_max.state)

        # motion off | extended not active
        self.motion_max.state = "Unlocked_Motion"
        with unittest.mock.patch.object(self.motion_max, "_motion_extended_configured", return_value=False):
            self.motion_max.motion_off()
        self.assertEqual("Unlocked_Wait", self.motion_max.state)

    def test_unlocked_motion_extended(self) -> None:
        """Test leaving transitions of Unlocked_MotionExtended state."""
        # timeout | brightness over threshold
        self.motion_max.state = "Unlocked_MotionExtended"
        with unittest.mock.patch.object(self.motion_max, "_brightness_over_threshold", return_value=True):
            self.motion_max.timeout_motion_extended()
            self.assertEqual("Unlocked_TooBright", self.motion_max.state)

        # timeout | brightness below threshold
        self.motion_max.state = "Unlocked_MotionExtended"
        with unittest.mock.patch.object(self.motion_max, "_brightness_over_threshold", return_value=False):
            self.motion_max.timeout_motion_extended()
            self.assertEqual("Unlocked_Wait", self.motion_max.state)

        # motion on
        self.motion_max.state = "Unlocked_MotionExtended"
        self.motion_max.motion_on()
        self.assertEqual("Unlocked_Motion", self.motion_max.state)

    def test_unlocked_too_bright(self) -> None:
        """Test leaving transitions of Unlocked_TooBright state."""
        # motion not active
        self.motion_max.state = "Unlocked_TooBright"
        with unittest.mock.patch.object(self.motion_max, "_raw_motion_active", return_value=False):
            self.motion_max.brightness_below_threshold()
        self.assertEqual("Unlocked_Wait", self.motion_max.state)

        # motion active
        self.motion_max.state = "Unlocked_TooBright"
        with unittest.mock.patch.object(self.motion_max, "_raw_motion_active", return_value=True):
            self.motion_max.brightness_below_threshold()
        self.assertEqual("Unlocked_Motion", self.motion_max.state)

    def test_check_brightness(self) -> None:
        """Test _check_brightness."""
        with (
            unittest.mock.patch.object(self.motion_max._hysteresis_switch, "get_output", return_value=True),
            unittest.mock.patch.object(self.motion_max, "brightness_over_threshold"),
            unittest.mock.patch.object(self.motion_max, "brightness_below_threshold"),
        ):
            self.motion_max._check_brightness()
            self.motion_max.brightness_over_threshold.assert_called_once()
            self.motion_max.brightness_below_threshold.assert_not_called()

        with (
            unittest.mock.patch.object(self.motion_max._hysteresis_switch, "get_output", return_value=False),
            unittest.mock.patch.object(self.motion_max, "brightness_over_threshold"),
            unittest.mock.patch.object(self.motion_max, "brightness_below_threshold"),
        ):
            self.motion_max._check_brightness()
            self.motion_max.brightness_over_threshold.assert_not_called()
            self.motion_max.brightness_below_threshold.assert_called_once()

    def test_cb_brightness_threshold_change(self) -> None:
        """Test _cb_threshold_change."""
        with unittest.mock.patch.object(self.motion_max._hysteresis_switch, "set_threshold_on"), unittest.mock.patch.object(self.motion_max, "_check_brightness"):
            tests.helper.oh_item.item_state_change_event("Unittest_Brightness_Threshold", 42)
            self.motion_max._hysteresis_switch.set_threshold_on.assert_called_once_with(42)
            self.motion_max._check_brightness.assert_called_once()

    def test_cb_motion_raw(self) -> None:
        """Test _cb_motion_raw."""
        with unittest.mock.patch.object(self.motion_max, "motion_on"), unittest.mock.patch.object(self.motion_max, "motion_off"):
            tests.helper.oh_item.item_state_change_event("Unittest_Motion_max_raw", "ON", "OFF")
            self.motion_max.motion_on.assert_called_once()
            self.motion_max.motion_off.assert_not_called()

        with unittest.mock.patch.object(self.motion_max, "motion_on"), unittest.mock.patch.object(self.motion_max, "motion_off"):
            tests.helper.oh_item.item_state_change_event("Unittest_Motion_max_raw", "OFF", "ON")
            self.motion_max.motion_on.assert_not_called()
            self.motion_max.motion_off.assert_called_once()

    def test_cb_brightness_change(self) -> None:
        """Test _cb_threshold_change."""
        with unittest.mock.patch.object(self.motion_max, "_check_brightness"):
            tests.helper.oh_item.item_state_change_event("Unittest_Brightness", 42)
            self.motion_max._check_brightness.assert_called_once()

    def test_cb_sleep(self) -> None:
        """Test _cb_sleep."""
        for state in habapp_rules.system.SleepState:
            with unittest.mock.patch.object(self.motion_max, "sleep_started"), unittest.mock.patch.object(self.motion_max, "sleep_end"):
                tests.helper.oh_item.item_state_change_event("Unittest_Sleep_state", state.value)
                if state == habapp_rules.system.SleepState.SLEEPING:
                    self.motion_max.sleep_started.assert_called_once()
                    self.motion_max.sleep_end.assert_not_called()

                elif state == habapp_rules.system.SleepState.AWAKE:
                    self.motion_max.sleep_started.assert_not_called()
                    self.motion_max.sleep_end.assert_called_once()

                else:
                    self.motion_max.sleep_started.assert_not_called()
                    self.motion_max.sleep_end.assert_not_called()



"""Common part for tests with simulated OpenHAB items."""

import threading
import unittest
import unittest.mock

import habapp_rules.core.state_machine_rule
import tests.helper.oh_item
import tests.helper.rule_runner


class TestCaseBase(unittest.TestCase):
    """Base class for tests with simulated OpenHAB items."""

    def setUp(self) -> None:
        """Setup test case."""
        self.send_command_mock_patcher = unittest.mock.patch("HABApp.openhab.items.base_item.send_command", new=tests.helper.oh_item.send_command)
        self.addCleanup(self.send_command_mock_patcher.stop)
        self.send_command_mock = self.send_command_mock_patcher.start()

        self.send_command_mock_patcher = unittest.mock.patch("HABApp.openhab.items.base_item.post_update", new=tests.helper.oh_item.set_state)
        self.addCleanup(self.send_command_mock_patcher.stop)
        self.send_command_mock = self.send_command_mock_patcher.start()

        self.item_exists_mock_patcher = unittest.mock.patch("HABApp.openhab.interface_sync.item_exists", return_value=True)
        self.addCleanup(self.item_exists_mock_patcher.stop)
        self.item_exists_mock = self.item_exists_mock_patcher.start()

        self._runner = tests.helper.rule_runner.SimpleRuleRunner()
        self._runner.set_up()

    def tearDown(self) -> None:
        """Tear down test case."""
        tests.helper.oh_item.remove_all_mocked_items()
        self._runner.tear_down()


class TestCaseBaseStateMachine(TestCaseBase):
    """Base class for tests with simulated OpenHAB items and state machines."""

    def setUp(self) -> None:
        """Setup tests."""
        TestCaseBase.setUp(self)

        self.transitions_timer_mock_patcher = unittest.mock.patch("transitions.extensions.states.Timer", spec=threading.Timer)
        self.addCleanup(self.transitions_timer_mock_patcher.stop)
        self.transitions_timer_mock = self.transitions_timer_mock_patcher.start()

        self.threading_timer_mock_patcher = unittest.mock.patch("threading.Timer", spec=threading.Timer)
        self.addCleanup(self.threading_timer_mock_patcher.stop)
        self.threading_timer_mock = self.threading_timer_mock_patcher.start()

        self.on_rule_removed_mock_patcher = unittest.mock.patch("habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed", spec=habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed)
        self.addCleanup(self.on_rule_removed_mock_patcher.stop)
        self.on_rule_removed_mock_patcher.start()

    def _get_state_names(self, states: dict, parent_state: str | None = None) -> list[str]:  # pragma: no cover
        """Helper function to get all state names (also nested states).

        Args:
            states: dict of all states or children states
            parent_state: name of parent state, only if it is a nested state machine

        Returns:
            list of all state names
        """
        state_names = []
        prefix = f"{parent_state}_" if parent_state else ""
        if parent_state:
            states = states["children"]

        for state in states:
            if "children" in state:
                state_names += self._get_state_names(state, state["name"])
            else:
                state_names.append(f"{prefix}{state['name']}")
        return state_names



"""Unit tests for habapp_rules helper."""

import collections
import time
import unittest
import unittest.mock

import HABApp
import whenever

import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import tests.helper.oh_item
import tests.helper.test_case_base


class TestHelperFunctions(tests.helper.test_case_base.TestCaseBase):
    """Tests for all helper functions."""

    def test_create_additional_item(self) -> None:
        """Test create additional item."""
        # check if item is created if NOT existing
        self.item_exists_mock.return_value = False
        TestCase = collections.namedtuple("TestCase", "item_type, name, label_input, label_call, groups")

        test_cases = [
            TestCase("Switch", "Item_name", "Some label", "Some label", None),
            TestCase("Switch", "Item_name", None, "Item name", None),
            TestCase("String", "Item_name", "Some label", "Some label", None),
            TestCase("String", "Item_name", "Some label", "Some label", None),
            TestCase("String", "Item_name", None, "Item name", None),
            TestCase("String", "Item_name", None, "Item name", ["test_group"]),
        ]

        with unittest.mock.patch("HABApp.openhab.interface_sync.create_item", spec=HABApp.openhab.interface_sync.create_item) as create_mock, unittest.mock.patch("HABApp.openhab.items.OpenhabItem.get_item"):
            for test_case in test_cases:
                create_mock.reset_mock()
                habapp_rules.core.helper.create_additional_item(test_case.name, test_case.item_type, test_case.label_input, test_case.groups)
                create_mock.assert_called_once_with(item_type=test_case.item_type, name=f"H_{test_case.name}", label=test_case.label_call, groups=test_case.groups)

        # check if item is NOT created if existing
        self.item_exists_mock.return_value = True
        with unittest.mock.patch("HABApp.openhab.interface_sync.create_item", spec=HABApp.openhab.interface_sync.create_item) as create_mock, unittest.mock.patch("HABApp.openhab.items.OpenhabItem.get_item"):
            habapp_rules.core.helper.create_additional_item("Name_of_Item", "Switch")
            create_mock.assert_not_called()

    def test_test_create_additional_item_exception(self) -> None:
        """Test exceptions of _create_additional_item."""
        self.item_exists_mock.return_value = False
        with unittest.mock.patch("HABApp.openhab.interface_sync.create_item", spec=HABApp.openhab.interface_sync.create_item, return_value=False), self.assertRaises(habapp_rules.core.exceptions.HabAppRulesError):
            habapp_rules.core.helper.create_additional_item("Name_of_Item", "Switch")

    def test_send_if_different(self) -> None:
        """Test send_if_different."""
        # item given
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number", 0)
        number_item = HABApp.openhab.items.NumberItem.get_item("Unittest_Number")

        habapp_rules.core.helper.send_if_different(number_item, 0)
        tests.helper.oh_item.assert_value("Unittest_Number", 0)

        habapp_rules.core.helper.send_if_different(number_item, 42)
        tests.helper.oh_item.assert_value("Unittest_Number", 42)

        # name given
        habapp_rules.core.helper.send_if_different("Unittest_Number", 42)
        tests.helper.oh_item.assert_value("Unittest_Number", 42)

        habapp_rules.core.helper.send_if_different("Unittest_Number", 84)
        tests.helper.oh_item.assert_value("Unittest_Number", 84)


class TestHelperWithItems(tests.helper.test_case_base.TestCaseBase):
    """Test helper functions with OpenHAB items."""

    def test_filter_updated_items(self) -> None:
        """Test filter_updated_items."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer", 0)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", "OFF")

        item_number = HABApp.openhab.items.NumberItem.get_item("Unittest_Number")
        item_dimmer = HABApp.openhab.items.DimmerItem.get_item("Unittest_Dimmer")
        item_switch = HABApp.openhab.items.SwitchItem.get_item("Unittest_Switch")

        # without filter
        result = habapp_rules.core.helper.filter_updated_items([item_number, item_dimmer, item_switch])
        self.assertListEqual([item_number, item_dimmer, item_switch], result)

        # with filter
        result = habapp_rules.core.helper.filter_updated_items([item_number, item_dimmer, item_switch], 60)
        self.assertListEqual([item_number, item_dimmer, item_switch], result)

        item_dimmer._last_update = HABApp.core.items.base_item.UpdatedTime("Unittest_Dimmer", whenever.Instant.from_timestamp(time.time() - 61))
        result = habapp_rules.core.helper.filter_updated_items([item_number, item_dimmer, item_switch], 60)
        self.assertListEqual([item_number, item_switch], result)



"""Config models for irrigation rules."""

import HABApp.openhab.items
import pydantic
import typing_extensions

import habapp_rules.core.exceptions
import habapp_rules.core.pydantic_base


class IrrigationItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for irrigation rules."""

    valve: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="valve item which will be switched")
    active: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to activate the rule")
    hour: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="start hour")
    minute: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="start minute")
    duration: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="duration in minutes")
    repetitions: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="number of repetitions")
    brake: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="time in minutes between repetitions")

    @pydantic.model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Validate model.

        Returns:
            validated model

        Raises:
            AssertionError: if 'repetitions' and 'brake' are not set together
        """
        if (self.repetitions is None) != (self.brake is None):
            msg = "If repeats item is given, also the brake item must be given!"
            raise AssertionError(msg)

        return self


class IrrigationConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for irrigation actors."""

    items: IrrigationItems = pydantic.Field(..., description="items for irrigation rule")
    parameter: None = None



"""Get type of day."""

import datetime

import holidays

_BY_HOLIDAYS = holidays.country_holidays("DE", "BY")


def is_weekend(offset_days: int = 0) -> bool:
    """Check if current (or offset day) is a weekend day.

    Args:
        offset_days: 0 means today, 1 tomorrow, -1 yesterday

    Returns:
        True if desired day is a weekend day
    """
    day_to_check = datetime.datetime.now() + datetime.timedelta(days=offset_days)
    return day_to_check.isoweekday() in {6, 7}


def is_holiday(offset_days: int = 0) -> bool:
    """Check if current (or offset day) is holiday.

    Args:
        offset_days: 0 means today, 1 tomorrow, -1 yesterday

    Returns:
        True if desired day is holiday
    """
    day_to_check = datetime.datetime.now() + datetime.timedelta(days=offset_days)
    return day_to_check in _BY_HOLIDAYS



"""Rule to detect whether it is summer or winter."""

import datetime
import logging
import statistics

import HABApp

import habapp_rules.common.hysteresis
import habapp_rules.core.logger
import habapp_rules.system.config.summer_winter

LOGGER = logging.getLogger(__name__)


class SummerWinterError(Exception):
    """Custom Exception for SummerWinter."""


class SummerWinter(HABApp.Rule):
    """Rule check if it is summer or winter."""

    def __init__(self, config: habapp_rules.system.config.summer_winter.SummerWinterConfig) -> None:
        """Init rule to update summer/winter item.

        Args:
            config: Config for summer/winter detection
        """
        self._config = config
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.summer.name)

        # set class variables
        self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(config.parameter.temperature_threshold, 0.5)
        self.__now = datetime.datetime.now()

        # run at init and every day at 23:00
        self.run.soon(self._cb_update_summer)
        self.run.at(self.run.trigger.time("23:00:00"), self._cb_update_summer)

        LOGGER.debug("Init of Summer / Winter successful")

    def __get_weighted_mean(self, days_in_past: int) -> float:
        """Get weighted mean temperature.

        The weighted mean temperature will be calculated according the following formula: (T07 + T14 + T22 + T22) / 4 where T07 is the temperature at 7:00 (and so on)
        It is possible to get the weighted temperature of today or of some days in the past -> defined by the days_in past attribute. If this method is called before 22:00 there will be an offset of one day.

        Args:
            days_in_past: if days in past is set to 0 it will return the mean of today. 1 will return the offset of yesterday

        Returns:
            the weighted mean according the formula in doc-string

        Raises:
            SummerWinterError: if there is not enough data for at least one evaluated hour.
        """
        day_offset = 0
        if self.__now.hour < 23:  # noqa: PLR2004
            day_offset = 1

        temperature_values = []
        for hour in [7, 14, 22]:
            start_time = datetime.datetime(self.__now.year, self.__now.month, self.__now.day, hour, 0) - datetime.timedelta(days=day_offset + days_in_past)
            end_time = start_time + datetime.timedelta(hours=1)
            persistence_data = self._config.items.outside_temperature.get_persistence_data(persistence=self._config.parameter.persistence_service, start_time=start_time, end_time=end_time)
            if not persistence_data.data:
                msg = f"No data for {start_time}"
                raise SummerWinterError(msg)
            temperature_values.append(next(iter(persistence_data.data.values())))
        return (sum(temperature_values) + temperature_values[2]) / 4

    def __is_summer(self) -> bool:
        """Check if it is summer (or winter).

        Returns:
            Returns True if it is summer.

        Raises:
            SummerWinterError: if summer/winter could not be detected
        """
        self.__now = datetime.datetime.now()
        values = []
        for day in range(self._config.parameter.days):
            try:
                values.append(self.__get_weighted_mean(day))
            except SummerWinterError:  # noqa: PERF203
                self._instance_logger.warning(f"Could not get mean value of day -{day}")

        if not values:
            msg = "Not enough data to detect summer/winter"
            raise SummerWinterError(msg)

        is_summer = self._hysteresis_switch.get_output(mean_value := statistics.mean(values))
        self._instance_logger.debug(f"Check Summer/Winter. values = {values} | mean = {mean_value} | summer = {is_summer}")
        return is_summer

    def _cb_update_summer(self) -> None:
        """Callback to update the summer item."""
        try:
            is_summer = self.__is_summer()
        except SummerWinterError:
            self._instance_logger.exception("Could not get summer / winter")
            return

        # get target state of summer
        target_value = "ON" if is_summer else "OFF"

        # send state
        if self._config.items.summer.value != target_value:
            self._instance_logger.info(f"Summer changed to {target_value}")
        self._config.items.summer.oh_send_command(target_value)

        # update last update item at every call
        if self._config.items.last_check is not None:
            self._config.items.last_check.oh_send_command(datetime.datetime.now())



"""Implementations of logical functions."""

import abc
import logging

import HABApp

import habapp_rules.common.config.logic
import habapp_rules.core.helper
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class _BinaryLogicBase(HABApp.Rule):
    """Base class for binary logical functions."""

    def __init__(self, config: habapp_rules.common.config.logic.BinaryLogicConfig) -> None:
        """Init a logical function.

        Args:
            config: Config for logical function

        Raises:
            TypeError: if unsupported item-type is given for output_name
        """
        HABApp.Rule.__init__(self)
        self._config = config
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

        if isinstance(self._config.items.output, HABApp.openhab.items.SwitchItem):
            # item type is Switch
            self._positive_state = "ON"
            self._negative_state = "OFF"
        else:
            # item type is Contact (validated by type of config)
            self._positive_state = "CLOSED"
            self._negative_state = "OPEN"

        for item in self._config.items.inputs:
            item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateUpdatedEventFilter())

        self._cb_input_event(None)
        input_names = [item.name for item in self._config.items.inputs]
        self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input items = {input_names}")

    @abc.abstractmethod
    def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
        """Callback, which is called if one of the input items had a state event.

        Args:
            event: item event of the updated item
        """

    def _set_output_state(self, output_state: str) -> None:
        """Set state to the output element.

        Args:
            output_state: state which will be set
        """
        if isinstance(self._config.items.output, HABApp.openhab.items.ContactItem):
            self._config.items.output.oh_post_update(output_state)
        else:
            habapp_rules.core.helper.send_if_different(self._config.items.output, output_state)


class And(_BinaryLogicBase):
    """Logical AND function.

    # Config:
    config = habapp_rules.common.config.logic.BinaryLogicConfig(
            items=habapp_rules.common.config.logic.BinaryLogicItems(
                    inputs=["Item_1", "Item_2"],
                    output="Item_result",
            )
    )

    # Rule init:
    habapp_rules.common.logic.And(config)
    """

    def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:  # noqa: ARG002
        """Callback, which is called if one of the input items had a state event.

        Args:
            event: item event of the updated item
        """
        output_state = self._positive_state if all(item.value == self._positive_state for item in self._config.items.inputs) else self._negative_state
        self._set_output_state(output_state)


class Or(_BinaryLogicBase):
    """Logical OR function.

    # Config:
    config = habapp_rules.common.config.logic.BinaryLogicConfig(
            items=habapp_rules.common.config.logic.BinaryLogicItems(
                    inputs=["Item_1", "Item_2"],
                    output="Item_result",
            )
    )

    # Rule init:
    habapp_rules.common.logic.Or(config)
    """

    def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:  # noqa: ARG002
        """Callback, which is called if one of the input items had a state event.

        Args:
            event: item event of the updated item
        """
        output_state = self._positive_state if any(item.value == self._positive_state for item in self._config.items.inputs) else self._negative_state
        self._set_output_state(output_state)


class _NumericLogicBase(HABApp.Rule):
    """Base class for numeric logical functions."""

    def __init__(self, config: habapp_rules.common.config.logic.NumericLogicConfig) -> None:
        """Init a logical function.

        Args:
            config: Config for logical function

        Raises:
            TypeError: if unsupported item-type is given for output_name
        """
        HABApp.Rule.__init__(self)
        self._config = config
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

        for item in self._config.items.inputs:
            item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._cb_input_event(None)
        input_names = [item.name for item in self._config.items.inputs]
        self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input items = {input_names}")

    def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:  # noqa: ARG002
        """Callback, which is called if one of the input items had a state event.

        Args:
            event: item event of the updated item
        """
        filtered_items = habapp_rules.core.helper.filter_updated_items(self._config.items.inputs, self._config.parameter.ignore_old_values_time)
        value = self._apply_numeric_logic([item.value for item in filtered_items if item is not None])

        if value is None:
            return

        self._set_output_state(value)

    @staticmethod
    @abc.abstractmethod
    def _apply_numeric_logic(input_values: list[float]) -> float:
        """Apply numeric logic.

        Args:
            input_values: input values

        Returns:
            value which fulfills the filter type
        """

    def _set_output_state(self, output_state: float) -> None:
        """Set state to the output element.

        Args:
            output_state: state which will be set
        """
        habapp_rules.core.helper.send_if_different(self._config.items.output, output_state)


class Min(_NumericLogicBase):
    """Logical Min function with filter for old / not updated items.

    # Config:
    config = habapp_rules.common.config.logic.NumericLogicConfig(
            items=habapp_rules.common.config.logic.NumericLogicItems(
                    inputs=["Item_1", "Item_2"],
                    output="Item_result",
            ),
            parameter=habapp_rules.common.config.logic.NumericLogicParameter(
                    ignore_old_values_time=600
            ),
    )

    # Rule init:
    habapp_rules.common.logic.Min(config)
    """

    @staticmethod
    def _apply_numeric_logic(input_values: list[float]) -> float:
        """Apply numeric logic.

        Args:
            input_values: input values

        Returns:
            min value of the given values
        """
        return HABApp.util.functions.min(input_values)


class Max(_NumericLogicBase):
    """Logical Max function with filter for old / not updated items.

    # Config:
    config = habapp_rules.common.config.logic.NumericLogicConfig(
            items=habapp_rules.common.config.logic.NumericLogicItems(
                    inputs=["Item_1", "Item_2"],
                    output="Item_result",
            ),
            parameter=habapp_rules.common.config.logic.NumericLogicParameter(
                    ignore_old_values_time=600
            ),
    )

    # Rule init:
    habapp_rules.common.logic.Max(config)
    """

    @staticmethod
    def _apply_numeric_logic(input_values: list[float]) -> float:
        """Apply numeric logic.

        Args:
            input_values: input values

        Returns:
            max value of the given values
        """
        return HABApp.util.functions.max(input_values)


class Sum(_NumericLogicBase):
    """Logical Sum function with filter for old / not updated items.

    # Config:
    config = habapp_rules.common.config.logic.NumericLogicConfig(
            items=habapp_rules.common.config.logic.NumericLogicItems(
                    inputs=["Item_1", "Item_2"],
                    output="Item_result",
            ),
            parameter=habapp_rules.common.config.logic.NumericLogicParameter(
                    ignore_old_values_time=600
            ),
    )

    # Rule init:
    habapp_rules.common.logic.Sum(config)
    """

    def __init__(self, config: habapp_rules.common.config.logic.NumericLogicConfig) -> None:
        """Init a logical function.

        Args:
            config: config for logical sum rule

        Raises:
            TypeError: if unsupported item-type is given for output_name
        """
        if isinstance(config.items.output, HABApp.openhab.items.DimmerItem):
            msg = f"Dimmer items can not be used for Sum function! Given output_name: {config.items.output}"
            raise TypeError(msg)

        _NumericLogicBase.__init__(self, config)

    @staticmethod
    def _apply_numeric_logic(input_values: list[float]) -> float:
        """Apply numeric logic.

        Args:
            input_values: input values

        Returns:
            min value of the given values
        """
        return sum(val for val in input_values if val is not None)


class InvertValue(HABApp.Rule):
    """Rule to update another item if the value of an item changed.

    # Config:
    config = habapp_rules.common.config.logic.InvertValueConfig(
            items=habapp_rules.common.config.logic.InvertValueItems(
                    input="Item_1",
                    output="Item_2",
            )
    )

    # Rule init:
    habapp_rules.common.logic.InvertValue(config)
    """

    def __init__(self, config: habapp_rules.common.config.logic.InvertValueConfig) -> None:
        """Init rule.

        Args:
            config: Config for invert value rule
        """
        HABApp.Rule.__init__(self)
        self._config = config
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

        self._config.items.input.listen_event(self._cb_input_value, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._cb_input_value(HABApp.openhab.events.ItemStateChangedEvent(self._config.items.input.name, self._config.items.input.value, None))
        self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input item = '{self._config.items.input.name}'")

    def _cb_input_value(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Set output, when input value changed.

        Args:
            event: event, which triggered this callback
        """
        if event.value is None:
            return

        output_value = -1 * event.value

        if (self._config.parameter.only_negative and output_value > 0) or (self._config.parameter.only_positive and output_value < 0):
            output_value = 0

        self._config.items.output.oh_send_command(output_value)



"""Test config models of logic rules."""

import HABApp

import habapp_rules.common.config.logic
import tests.helper.oh_item
import tests.helper.test_case_base


class TestBinaryLogicItems(tests.helper.test_case_base.TestCaseBase):
    """Test BinaryLogicItems."""

    def tests_model_validator(self) -> None:
        """Tests model_validator."""
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Input_Switch", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Input_Contact", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_Switch", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.ContactItem, "Unittest_Output_Contact", None)

        # input and output items are the same
        habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Input_Switch"], output="Unittest_Output_Switch")

        habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Input_Contact"], output="Unittest_Output_Contact")

        # input and output items are different
        with self.assertRaises(TypeError):
            habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Input_Switch"], output="Unittest_Output_Contact")

        with self.assertRaises(TypeError):
            habapp_rules.common.config.logic.BinaryLogicItems(inputs=["Unittest_Input_Contact"], output="Unittest_Output_Switch")



"""Config models for motion rules."""

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class MotionItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for motion."""

    motion_raw: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="unfiltered motion item")
    motion_filtered: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="filtered motion item")
    brightness: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="brightness item")
    brightness_threshold: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="brightness threshold item")
    lock: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="lock item")
    sleep_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleep state item")
    state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="state item")


class MotionParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for motion."""

    extended_motion_time: int = pydantic.Field(5, description="extended motion time in seconds")
    brightness_threshold: float | None = pydantic.Field(None, description="brightness threshold value")
    post_sleep_lock_time: int = pydantic.Field(10, description="post sleep lock time in seconds")


class MotionConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for motion."""

    items: MotionItems = pydantic.Field(..., description="items for motion")
    parameter: MotionParameter = pydantic.Field(MotionParameter(), description="parameter for motion")



"""Test version."""

import platform

import HABApp.openhab.items

import habapp_rules
import habapp_rules.core.version
import tests.helper.oh_item
import tests.helper.test_case_base


class TestSetVersions(tests.helper.test_case_base.TestCaseBase):
    """Test for SetVersions."""

    def setUp(self) -> None:
        """Set up test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_habapp_version", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_habapp_rules_version", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_python_version", None)

        habapp_rules.core.version.SetVersions()

    def test_version_values(self) -> None:
        """Test if versions were set correctly."""
        tests.helper.oh_item.assert_value("H_habapp_version", HABApp.__version__)
        tests.helper.oh_item.assert_value("H_habapp_rules_version", habapp_rules.__version__)
        tests.helper.oh_item.assert_value("H_python_version", platform.python_version())



"""Test power rules."""

import collections
import unittest.mock

import HABApp.rule.rule
import HABApp.rule.scheduler.job_ctrl

import habapp_rules.core.exceptions
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.current_switch
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer
from habapp_rules.sensors.config.current_switch import CurrentSwitchConfig, CurrentSwitchItems, CurrentSwitchParameter


class TestCurrentSwitch(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing CurrentSwitch rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Current", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_extended", None)

        self._rule_1 = habapp_rules.sensors.current_switch.CurrentSwitch(
            CurrentSwitchConfig(
                items=CurrentSwitchItems(
                    current="Unittest_Current",
                    switch="Unittest_Switch_1",
                )
            )
        )

        self._rule_2 = habapp_rules.sensors.current_switch.CurrentSwitch(
            CurrentSwitchConfig(
                items=CurrentSwitchItems(
                    current="Unittest_Current",
                    switch="Unittest_Switch_2",
                ),
                parameter=CurrentSwitchParameter(threshold=1),
            )
        )

        self._rule_extended = habapp_rules.sensors.current_switch.CurrentSwitch(
            CurrentSwitchConfig(
                items=CurrentSwitchItems(
                    current="Unittest_Current",
                    switch="Unittest_Switch_extended",
                ),
                parameter=CurrentSwitchParameter(extended_time=60),
            )
        )

    def test_init(self) -> None:
        """Test __init__."""
        tests.helper.oh_item.assert_value("Unittest_Switch_1", None)
        tests.helper.oh_item.assert_value("Unittest_Switch_2", None)
        tests.helper.oh_item.assert_value("Unittest_Switch_extended", None)

        self.assertIsNone(self._rule_1._extended_countdown)
        self.assertIsNone(self._rule_2._extended_countdown)
        self.assertIsInstance(self._rule_extended._extended_countdown, HABApp.rule.scheduler.job_ctrl.CountdownJobControl)
        self.assertIsNone(self._rule_extended._extended_countdown.next_run_datetime)

    def test_current_changed_without_extended_time(self) -> None:
        """Test current changed without extended time."""
        TestCase = collections.namedtuple("TestCase", "current, expected_1, expected_2")

        test_cases = [
            TestCase(0, "OFF", "OFF"),
            TestCase(0.2, "OFF", "OFF"),
            TestCase(0.201, "ON", "OFF"),
            TestCase(1, "ON", "OFF"),
            TestCase(1.001, "ON", "ON"),
            TestCase(1.001, "ON", "ON"),
            TestCase(1, "ON", "OFF"),
            TestCase(0.200, "OFF", "OFF"),
            TestCase(0, "OFF", "OFF"),
            TestCase(-10000, "OFF", "OFF"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_Current", test_case.current)

                tests.helper.oh_item.assert_value("Unittest_Switch_1", test_case.expected_1)
                tests.helper.oh_item.assert_value("Unittest_Switch_2", test_case.expected_2)

    def test_current_changed_with_extended_time(self) -> None:
        """Test current changed with extended time."""
        with unittest.mock.patch.object(self._rule_extended, "_extended_countdown") as countdown_mock:
            # below threshold
            tests.helper.oh_item.item_state_change_event("Unittest_Current", 0.1)
            tests.helper.oh_item.assert_value("Unittest_Switch_extended", None)
            countdown_mock.stop.assert_not_called()
            countdown_mock.reset.assert_not_called()

            # above threshold
            tests.helper.oh_item.item_state_change_event("Unittest_Current", 0.3)
            tests.helper.oh_item.assert_value("Unittest_Switch_extended", "ON")
            countdown_mock.stop.assert_called_once()
            countdown_mock.reset.assert_not_called()

            # below threshold
            countdown_mock.stop.reset_mock()
            tests.helper.oh_item.item_state_change_event("Unittest_Current", 0.1)
            tests.helper.oh_item.assert_value("Unittest_Switch_extended", "ON")
            countdown_mock.stop.assert_not_called()
            countdown_mock.reset.assert_called_once()



"""Rules to manage shading objects."""

import abc
import datetime
import logging
import time
import typing

import HABApp

import habapp_rules.actors.config.shading
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)

HAND_IGNORE_TIME = 1.5


class _ShadingBase(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Base class for shading objects."""

    states: typing.ClassVar = [
        {"name": "WindAlarm"},
        {"name": "Manual"},
        {"name": "Hand", "timeout": 20 * 3600, "on_timeout": "_auto_hand_timeout"},
        {
            "name": "Auto",
            "initial": "Init",
            "children": [
                {"name": "Init"},
                {"name": "Open"},
                {"name": "DoorOpen", "initial": "Open", "children": [{"name": "Open"}, {"name": "PostOpen", "timeout": 5 * 60, "on_timeout": "_timeout_post_door_open"}]},
                {"name": "NightClose"},
                {"name": "SleepingClose"},
                {"name": "SunProtection"},
            ],
        },
    ]

    trans: typing.ClassVar = [
        # wind alarm
        {"trigger": "_wind_alarm_on", "source": ["Auto", "Hand", "Manual"], "dest": "WindAlarm"},
        {"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Manual", "conditions": "_manual_active"},
        {"trigger": "_wind_alarm_off", "source": "WindAlarm", "dest": "Auto", "unless": "_manual_active"},
        # manual
        {"trigger": "_manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
        {"trigger": "_manual_off", "source": "Manual", "dest": "Auto"},
        # hand
        {"trigger": "_hand_command", "source": ["Auto"], "dest": "Hand"},
        {"trigger": "_auto_hand_timeout", "source": "Hand", "dest": "Auto"},
        # sun
        {"trigger": "_sun_on", "source": "Auto_Open", "dest": "Auto_SunProtection"},
        {"trigger": "_sun_off", "source": "Auto_SunProtection", "dest": "Auto_Open"},
        # sleep
        {"trigger": "_sleep_started", "source": ["Auto_Open", "Auto_NightClose", "Auto_SunProtection", "Auto_DoorOpen"], "dest": "Auto_SleepingClose"},
        {"trigger": "_sleep_started", "source": "Hand", "dest": "Auto"},
        {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
        {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_NightClose", "conditions": ["_night_active_and_configured"]},
        {"trigger": "_sleep_stopped", "source": "Auto_SleepingClose", "dest": "Auto_Open", "unless": ["_night_active_and_configured", "_sun_protection_active_and_configured"]},
        # door
        {"trigger": "_door_open", "source": ["Auto_NightClose", "Auto_SunProtection", "Auto_SleepingClose", "Auto_Open"], "dest": "Auto_DoorOpen"},
        {"trigger": "_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_DoorOpen_Open"},
        {"trigger": "_door_closed", "source": "Auto_DoorOpen_Open", "dest": "Auto_DoorOpen_PostOpen"},
        {"trigger": "_timeout_post_door_open", "source": "Auto_DoorOpen_PostOpen", "dest": "Auto_Init"},
        # night close
        {"trigger": "_night_started", "source": ["Auto_Open", "Auto_SunProtection"], "dest": "Auto_NightClose", "conditions": "_night_active_and_configured"},
        {"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_SunProtection", "conditions": "_sun_protection_active_and_configured"},
        {"trigger": "_night_stopped", "source": "Auto_NightClose", "dest": "Auto_Open", "unless": ["_sun_protection_active_and_configured"]},
    ]
    _state_observer_pos: habapp_rules.actors.state_observer.StateObserverRollerShutter | habapp_rules.actors.state_observer.StateObserverDimmer

    def __init__(self, config: habapp_rules.actors.config.shading.ShadingConfig) -> None:
        """Init of _ShadingBase.

        Args:
            config: shading config

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationException: if given config / items are not valid
        """
        self._config = config
        self._set_shading_state_timestamp = 0

        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.shading_position.name)

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()
        self._apply_config()

        self._position_before = habapp_rules.actors.config.shading.ShadingPosition(self._config.items.shading_position.value)

        if isinstance(self._config.items.shading_position, HABApp.openhab.items.rollershutter_item.RollershutterItem):
            self._state_observer_pos = habapp_rules.actors.state_observer.StateObserverRollerShutter(
                self._config.items.shading_position.name, self._cb_hand, [item.name for item in self._config.items.shading_position_control], [item.name for item in self._config.items.shading_position_group], self._config.parameter.value_tolerance
            )
        else:
            # self._config.items.shading_position is instance of HABApp.openhab.items.dimmer_item.DimmerItem
            self._state_observer_pos = habapp_rules.actors.state_observer.StateObserverDimmer(
                self._config.items.shading_position.name,
                self._cb_hand,
                self._cb_hand,
                self._cb_hand,
                [item.name for item in self._config.items.shading_position_control],
                [item.name for item in self._config.items.shading_position_group],
                self._config.parameter.value_tolerance,
            )

        # callbacks
        self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.wind_alarm is not None:
            self._config.items.wind_alarm.listen_event(self._cb_wind_alarm, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.sun_protection is not None:
            self._config.items.sun_protection.listen_event(self._cb_sun, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.sleeping_state is not None:
            self._config.items.sleeping_state.listen_event(self._cb_sleep_state, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.night is not None:
            self._config.items.night.listen_event(self._cb_night, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.door is not None:
            self._config.items.door.listen_event(self._cb_door, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._update_openhab_state()

    def _apply_config(self) -> None:
        """Apply config to state machine."""
        # set timeouts
        self.state_machine.states["Auto"].states["DoorOpen"].states["PostOpen"].timeout = self._config.parameter.door_post_time
        self.state_machine.states["Manual"].timeout = self._config.parameter.manual_timeout

    def _get_initial_state(self, default_value: str = "") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.wind_alarm is not None and self._config.items.wind_alarm.is_on():
            return "WindAlarm"
        if self._config.items.manual.is_on():
            return "Manual"
        if self._config.items.door is not None and self._config.items.door.is_open():  # self._item_door.is_open():
            return "Auto_DoorOpen_Open"
        if self._config.items.sleeping_state is not None and self._config.items.sleeping_state.value in {habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value}:
            return "Auto_SleepingClose"
        if self._config.items.night is not None and self._config.items.night.is_on() and self._night_active_and_configured():
            return "Auto_NightClose"
        if self._sun_protection_active_and_configured():
            return "Auto_SunProtection"
        return "Auto_Open"

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_shading_state()

            if self._config.items.hand_manual_is_active_feedback is not None:
                self._config.items.hand_manual_is_active_feedback.oh_post_update("ON" if self.state in {"Manual", "Hand"} else "OFF")

            self._previous_state = self.state

    def _set_shading_state(self) -> None:
        """Set shading state."""
        if self._previous_state is None:
            # don't change value if called during init (_previous_state == None)
            return

        self._set_shading_state_timestamp = time.time()
        self._apply_target_position(self._get_target_position())

    @abc.abstractmethod
    def _apply_target_position(self, target_position: habapp_rules.actors.config.shading.ShadingPosition) -> None:
        """Apply target position by sending it via the observer(s).

        Args:
            target_position: target position of the shading object
        """

    def _get_target_position(self) -> habapp_rules.actors.config.shading.ShadingPosition | None:  # noqa: C901
        """Get target position for shading object.

        Returns:
            target shading position
        """
        if self.state in {"Hand", "Manual"}:
            if self._previous_state == "WindAlarm":
                return self._position_before
            return None

        if self.state == "WindAlarm":
            return self._config.parameter.pos_wind_alarm

        if self.state == "Auto_Open":
            return self._config.parameter.pos_auto_open

        if self.state == "Auto_SunProtection":
            return self._config.parameter.pos_sun_protection

        if self.state == "Auto_SleepingClose":
            if self._config.items.night is None:
                return self._config.parameter.pos_sleeping_night
            return self._config.parameter.pos_sleeping_night if self._config.items.night.is_on() else self._config.parameter.pos_sleeping_day

        if self.state == "Auto_NightClose":
            if self._config.items.summer is not None and self._config.items.summer.is_on():
                return self._config.parameter.pos_night_close_summer
            return self._config.parameter.pos_night_close_winter

        if self.state == "Auto_DoorOpen_Open":
            return self._config.parameter.pos_door_open

        return None

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Is called on entering of init state."""
        self._set_initial_state()

    def on_exit_Manual(self) -> None:  # noqa: N802
        """Is called if state Manual is left."""
        self._set_position_before()

    def on_exit_Hand(self) -> None:  # noqa: N802
        """Is called if state Hand is left."""
        self._set_position_before()

    def _set_position_before(self) -> None:
        """Set / save position before manual state is entered. This is used to restore the previous position."""
        self._position_before = habapp_rules.actors.config.shading.ShadingPosition(self._config.items.shading_position.value)

    def _manual_active(self) -> bool:
        """Check if manual is active.

        Returns:
            True if night is active
        """
        return self._config.items.manual.is_on()

    def _sun_protection_active_and_configured(self) -> bool:
        """Check if sun protection is active.

        Returns:
            True if night is active
        """
        return self._config.items.sun_protection is not None and self._config.items.sun_protection.is_on() and self._config.parameter.pos_sun_protection is not None

    def _night_active_and_configured(self) -> bool:
        """Check if night is active and configured.

        Returns:
            True if night is active
        """
        night_config = self._config.parameter.pos_night_close_summer if self._config.items.summer is not None and self._config.items.summer.is_on() else self._config.parameter.pos_night_close_winter
        return self._config.items.night is not None and self._config.items.night.is_on() and night_config is not None

    def _cb_hand(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if a external control was detected.

        Args:
            event: original trigger event
        """
        if time.time() - self._set_shading_state_timestamp > HAND_IGNORE_TIME:
            # ignore hand commands one second after this rule triggered a position change
            self._instance_logger.debug(f"Detected hand command. The event was {event}")
            self._hand_command()
        else:
            self._instance_logger.debug(f"Detected hand command, ignoring it. The event was {event}")

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if manual mode changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._manual_on()
        else:
            self._manual_off()

    def _cb_wind_alarm(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if wind alarm changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._wind_alarm_on()
        else:
            self._wind_alarm_off()

    def _cb_sun(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if sun state changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._sun_on()
        else:
            self._sun_off()

    def _cb_sleep_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if sleeping state changed.

        Args:
            event: original trigger event
        """
        if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
            self._sleep_started()
        elif event.value == habapp_rules.system.SleepState.POST_SLEEPING.value:
            self._sleep_stopped()

    def _cb_night(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if night / dark state changed.

        Args:
            event: original trigger event
        """
        if self.state == "Auto_SleepingClose":
            target_position = self._config.parameter.pos_sleeping_night if event.value == "ON" else self._config.parameter.pos_sleeping_day
            self._apply_target_position(target_position)

        if event.value == "ON":
            self._night_started()
        else:
            self._night_stopped()

    def _cb_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if door state changed.

        Args:
            event: original trigger event
        """
        if event.value == "OPEN":
            self._door_open()
        else:
            self._door_closed()


class Shutter(_ShadingBase):
    """Rules class to manage a normal shutters (or curtains).

    # KNX-things:
    Thing device KNX_Shading "KNX OpenHAB dimmer observer"{
        Type dimmer                 : shading_position          "Shading position"          [ position="5.001:4/1/12+<4/1/15" ]
        Type dimmer-control         : shading_position_ctr      "Shading position ctr"      [ position="5.001:4/1/12+<4/1/15" ]
        Type dimmer-control         : shading_group_all_ctr     "Shading all ctr"           [ position="5.001:4/1/112+<4/1/115""]
        Type switch-control         : shading_hand_manual_ctr   "Shading hand / manual"     [ ga="4/1/20" ]
    }

    # Items:
    Rollershutter    shading_position       "Shading position"                  <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position"}
    Rollershutter    shading_position_ctr   "Shading position ctr"              <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position_ctr"}
    Dimmer           shading_slat           "Shading slat"                      <slat>              {channel="knx:device:bridge:KNX_Shading:shading_slat"}
    Switch           shading_manual         "Shading manual"
    Rollershutter    shading_all_ctr        "Shading all ctr"                   <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_group_all_ctr"}
    Switch           shading_hand_manual    "Shading in Hand / Manual state"                        {channel="knx:device:bridge:KNX_Shading:shading_hand_manual_ctr"}

    # Config:
    config = habapp_rules.actors.config.shading.ShadingConfig(
            items = habapp_rules.actors.config.shading.ShadingItems(
                    shading_position="shading_position",
                    shading_position_control=["shading_position_ctr", "shading_all_ctr"],
                    slat="shading_slat",
                    manual="shading_manual",
                    wind_alarm="I99_99_WindAlarm",
                    sun_protection="I99_99_SunProtection",
                    sleeping_state="I99_99_Sleeping_State",
                    night="I99_99_Night",
                    door="I99_99_Door",
                    summer="I99_99_Summer",
                    hand_manual_is_active_feedback="shading_hand_manual"
            )
    )

    # Rule init:
    habapp_rules.actors.shading.Shutter(config)
    """

    def __init__(self, config: habapp_rules.actors.config.shading.ShadingConfig) -> None:
        """Init of Raffstore object.

        Args:
            config: shading config
        """
        _ShadingBase.__init__(self, config)

        self._instance_logger.debug(self.get_initial_log_message())

    def _apply_target_position(self, target_position: habapp_rules.actors.config.shading.ShadingPosition) -> None:
        """Apply target position by sending it via the observer(s).

        Args:
            target_position: target position of the shading object
        """
        if target_position is None:
            return

        if (position := target_position.position) is not None:
            self._state_observer_pos.send_command(position)
            self._instance_logger.debug(f"set position {target_position.position}")


class Raffstore(_ShadingBase):
    """Rules class to manage a raffstore.

    # KNX-things:
    Thing device KNX_Shading "KNX OpenHAB dimmer observer"{
        Type rollershutter          : shading_position          "Shading position"          [ upDown="4/1/10", stopMove="4/1/11", position="5.001:4/1/12+<4/1/15" ]
        Type rollershutter-control  : shading_position_ctr      "Shading position ctr"      [ upDown="4/1/10", stopMove="4/1/11" ]
        Type dimmer                 : shading_slat              "Shading slat"              [ position="5.001:4/1/13+<4/1/16" ]
        Type rollershutter-control  : shading_group_all_ctr     "Shading all ctr"           [ upDown="4/1/110", stopMove="4/1/111"]
        Type switch-control         : shading_hand_manual_ctr   "Shading hand / manual"     [ ga="4/1/20" ]
    }

    # Items:
    Rollershutter    shading_position       "Shading position"                  <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position"}
    Rollershutter    shading_position_ctr   "Shading position ctr"              <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_position_ctr"}
    Dimmer           shading_slat           "Shading slat"                      <slat>              {channel="knx:device:bridge:KNX_Shading:shading_slat"}
    Switch           shading_manual         "Shading manual"
    Rollershutter    shading_all_ctr        "Shading all ctr"                   <rollershutter>     {channel="knx:device:bridge:KNX_Shading:shading_group_all_ctr"}
    Switch           shading_hand_manual    "Shading in Hand / Manual state"                        {channel="knx:device:bridge:KNX_Shading:shading_hand_manual_ctr"}

    # Config:
    config = habapp_rules.actors.config.shading.ShadingConfig(
            items = habapp_rules.actors.config.shading.ShadingItems(
                    shading_position="shading_position",
                    shading_position_control=["shading_position_ctr", "shading_all_ctr"],
                    manual="shading_manual",
                    wind_alarm="I99_99_WindAlarm",
                    sun_protection="I99_99_SunProtection",
                    sleeping_state="I99_99_Sleeping_State",
                    night="I99_99_Night",
                    door="I99_99_Door",
                    summer="I99_99_Summer",
                    hand_manual_is_active_feedback="shading_hand_manual"
            )
    )

    # Rule init:
    habapp_rules.actors.shading.Raffstore(config)
    """

    def __init__(self, config: habapp_rules.actors.config.shading.ShadingConfig) -> None:
        """Init of Raffstore object.

        Args:
            config: shading config

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if the correct items are given for sun protection mode
        """
        # check if the correct items are given for sun protection mode
        if (config.items.sun_protection is None) != (config.items.sun_protection_slat is None):
            msg = "Ether items.sun_protection AND items.sun_protection_slat item must be given or None of them."
            raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)
        if config.items.slat is None:
            msg = "Item for setting the slat value must be given."
            raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

        _ShadingBase.__init__(self, config)

        self._state_observer_slat = habapp_rules.actors.state_observer.StateObserverSlat(config.items.slat.name, self._cb_hand, config.parameter.value_tolerance)

        # init items
        self.__verify_items()

        # callbacks
        if self._config.items.sun_protection_slat is not None:
            self._config.items.sun_protection_slat.listen_event(self._cb_slat_target, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._instance_logger.debug(self.get_initial_log_message())

    def __verify_items(self) -> None:
        """Check if given items are valid.

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if given items are not valid
        """
        # check type of rollershutter item
        if not isinstance(self._config.items.shading_position, HABApp.openhab.items.rollershutter_item.RollershutterItem):
            msg = f"The shading position item must be of type RollershutterItem. Given: {type(self._config.items.shading_position)}"
            raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)

    def _get_target_position(self) -> habapp_rules.actors.config.shading.ShadingPosition | None:
        """Get target position for shading object(s).

        Returns:
            target shading position
        """
        target_position = super()._get_target_position()

        if self.state == "Auto_SunProtection" and target_position is not None:
            target_position.slat = self._config.items.sun_protection_slat.value

        return target_position

    def _apply_target_position(self, target_position: habapp_rules.actors.config.shading.ShadingPosition) -> None:
        """Apply target position by sending it via the observer(s).

        Args:
            target_position: target position of the shading object
        """
        if target_position is None:
            return

        if (position := target_position.position) is not None:
            self._state_observer_pos.send_command(position)

        if (slat := target_position.slat) is not None:
            self._state_observer_slat.send_command(slat)

        if any(pos is not None for pos in (position, slat)):
            self._instance_logger.debug(f"set position {target_position}")

    def _set_position_before(self) -> None:
        """Set / save position before manual state is entered. This is used to restore the previous position."""
        self._position_before = habapp_rules.actors.config.shading.ShadingPosition(self._config.items.shading_position.value, self._config.items.slat.value)

    def _cb_slat_target(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the target slat value changed.

        Args:
            event: original trigger event
        """
        if self.state == "Auto_SunProtection":
            self._state_observer_slat.send_command(event.value)


class ResetAllManualHand(HABApp.Rule):
    """Clear the state hand / manual state of all shading.

    # Items:
    Switch           clear_hand_manual         "Clear Hand / Manual state of all shading objects"

    # Config
    config = habapp_rules.actors.config.shading.ResetAllManualHandConfig(
            items=habapp_rules.actors.config.shading.ResetAllManualHandItems(
                    reset_manual_hand="clear_hand_manual"
            )
    )

    # Rule init:
    habapp_rules.actors.shading.ResetAllManualHand(config)
    """

    def __init__(self, config: habapp_rules.actors.config.shading.ResetAllManualHandConfig) -> None:
        """Init of reset class.

        Args:
            config: config for reset all manual / hand rule
        """
        self._config = config
        HABApp.Rule.__init__(self)

        self._config.items.reset_manual_hand.listen_event(self._cb_reset_all, HABApp.openhab.events.ItemStateUpdatedEventFilter())

    def __get_shading_objects(self) -> list[_ShadingBase]:
        """Get all shading objects.

        Returns:
            list of shading objects
        """
        if self._config.parameter.shading_objects:
            return self._config.parameter.shading_objects
        return [rule for rule in self.get_rule(None) if issubclass(rule.__class__, _ShadingBase)]

    def _cb_reset_all(self, event: HABApp.openhab.events.ItemCommandEvent) -> None:
        """Callback which is called if reset is requested.

        Args:
            event: trigger event
        """
        if event.value == "OFF":
            return

        for shading_object in self.__get_shading_objects():
            state = shading_object.state
            manual_item = shading_object._config.items.manual  # noqa: SLF001

            if state == "Manual":
                manual_item.oh_send_command("OFF")

            elif state == "Hand":
                manual_item.oh_send_command("ON")
                manual_item.oh_send_command("OFF")

        self._config.items.reset_manual_hand.oh_send_command("OFF")


class SlatValueSun(HABApp.Rule):
    """Rules class to get slat value depending on sun elevation.

    # Items:
    Number    elevation             "Sun elevation"         <sun>     {channel="astro...}
    Number    sun_protection_slat   "Slat value"            <slat>

    # Config
    config = habapp_rules.actors.config.shading.SlatValueConfig(
            items=habapp_rules.actors.config.shading.SlatValueItems(
                    sun_elevation="elevation",
                    slat_value="sun_protection_slat",
                    summer="I99_99_Summer",
            )
    )

    # Rule init:
    habapp_rules.actors.shading.SlatValueSun(config)
    """

    def __init__(self, config: habapp_rules.actors.config.shading.SlatValueConfig) -> None:
        """Init SlatValueSun.

        Args:
            config: configuration of slat value

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
        """
        self._config = config
        HABApp.Rule.__init__(self)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.slat_value.name)

        # slat characteristics
        self._slat_characteristic_active = self._config.parameter.elevation_slat_characteristic_summer if self._config.items.summer is not None and self._config.items.summer.is_on() else self._config.parameter.elevation_slat_characteristic

        # callbacks
        self._config.items.sun_elevation.listen_event(self._cb_elevation, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.summer is not None:
            self._config.items.summer.listen_event(self._cb_summer_winter, HABApp.openhab.events.ItemStateChangedEventFilter())
        self.run.soon(self.__send_slat_value)

        self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with name '{self.rule_name}' was successful.")

    def __get_slat_value(self, elevation: float) -> float:
        """Get slat value for given elevation.

        Args:
            elevation: elevation of the sun

        Returns:
            slat value
        """
        if elevation >= self._slat_characteristic_active[-1].elevation:
            return self._slat_characteristic_active[-1].slat_value
        if elevation < self._slat_characteristic_active[0].elevation:
            return self._slat_characteristic_active[0].slat_value

        # no cover because of loop does not finish, but is handled with the two if statements above
        return next(config for idx, config in enumerate(self._slat_characteristic_active) if config.elevation <= elevation < self._slat_characteristic_active[idx + 1].elevation).slat_value  # pragma: no cover

    def __send_slat_value(self) -> None:
        """Send slat value to OpenHAB item."""
        if self._config.items.sun_elevation.value is None:
            return
        slat_value = self.__get_slat_value(self._config.items.sun_elevation.value)

        if self._config.items.slat_value.value != slat_value:
            self._config.items.slat_value.oh_send_command(slat_value)

    def _cb_elevation(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        """Callback which is called if sun elevation changed.

        Args:
            event: elevation event
        """
        self.__send_slat_value()

    def _cb_summer_winter(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is called if summer / winter changed.

        Args:
            event: summer / winter event
        """
        self._slat_characteristic_active = self._config.parameter.elevation_slat_characteristic_summer if event.value == "ON" else self._config.parameter.elevation_slat_characteristic
        self.__send_slat_value()


class ReferenceRun(HABApp.Rule):
    """Rule to trigger a reference run for blinds every month.

    # Items:
    Switch      trigger_run         "trigger reference run"
    DateTime    last_run            "last run"
    String      presence_state      "Presence state"

    # Config
    config = habapp_rules.actors.config.shading.ReferenceRunConfig(
            items=habapp_rules.actors.config.shading.ReferenceRunItems(
                trigger_run="trigger_run",
                presence_state="presence_state",
                last_run="last_run",
            )
    )

    # Rule init:
    habapp_rules.actors.shading.ReferenceRun(config)
    """

    def __init__(self, config: habapp_rules.actors.config.shading.ReferenceRunConfig) -> None:
        """Init ReferenceRun.

        Args:
            config: configuration of reference run rule
        """
        self._config = config
        HABApp.Rule.__init__(self)

        self._config.items.presence_state.listen_event(self._cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _cb_presence_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is called if presence state changed.

        Args:
            event: presence state event
        """
        if event.value == habapp_rules.system.PresenceState.ABSENCE.value:
            last_run = self._config.items.last_run.value or datetime.datetime.min
            current_time = datetime.datetime.now()

            if last_run.year < current_time.year or last_run.month < current_time.month:
                self._config.items.trigger_run.oh_send_command("ON")
                self._config.items.last_run.oh_send_command(current_time)



"""Config models for current switch rules."""

import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class CurrentSwitchItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for current switch the rule."""

    current: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item which measures the current")
    switch: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which should be switched on, if the current is above")


class CurrentSwitchParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for current switch the rules."""

    threshold: float = pydantic.Field(0.2, description="threshold for switching on")
    extended_time: float = pydantic.Field(0, description="extended time in seconds, if current is below threshold")


class CurrentSwitchConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config models for current switch the rule."""

    items: CurrentSwitchItems = pydantic.Field(..., description="items for current switch rules")
    parameter: CurrentSwitchParameter = pydantic.Field(CurrentSwitchParameter(), description="parameter for current switch rules")



import asyncio
from types import TracebackType

import HABApp
import HABApp.core.lib.exceptions.format
import HABApp.rule.rule as rule_module
import HABApp.rule.scheduler.job_builder as job_builder_module
from astral import Observer
from eascheduler.producers import prod_sun as prod_sun_module
from HABApp.core.asyncio import async_context
from HABApp.core.internals import EventBus, ItemRegistry, setup_internals
from HABApp.core.internals.proxy import ConstProxyObj
from HABApp.core.internals.wrapped_function import wrapped_thread, wrapper
from HABApp.core.internals.wrapped_function.wrapped_thread import WrappedThreadFunction
from HABApp.core.lib.exceptions.format import fallback_format
from HABApp.rule.rule_hook import HABAppRuleHook
from HABApp.runtime import Runtime
from pytest import MonkeyPatch


def suggest_rule_name(obj: object) -> str:
    return f"TestRule.{obj.__class__.__name__}"


class SyncScheduler:
    ALL = []

    def __init__(self, event_loop=None, enabled=True) -> None:
        SyncScheduler.ALL.append(self)
        self.jobs = []

    def add_job(self, job) -> None:
        self.jobs.append(job)

    def remove_job(self, job) -> None:
        if job in self.jobs:
            self.jobs.remove(job)

    def update_job(self, job) -> None:
        self.remove_job(job)
        self.add_job(job)

    def remove_all(self) -> None:
        self.jobs.clear()

    def set_enabled(self, enabled: bool) -> None:
        pass


class DummyRuntime(Runtime):
    def __init__(self) -> None:
        pass


def raising_fallback_format(e: Exception, existing_traceback: list[str]) -> list[str]:
    traceback = fallback_format(e, existing_traceback)
    traceback = traceback
    raise


class SimpleRuleRunner:
    def __init__(self) -> None:
        self.loaded_rules = []

        self.monkeypatch = MonkeyPatch()
        self.restore = []
        self.ctx = asyncio.Future()

    def submit(self, callback, *args, **kwargs) -> None:
        # This executes the callback so we can not ignore exceptions
        callback(*args, **kwargs)

    def set_up(self) -> None:
        # ensure that we call setup only once!
        assert isinstance(HABApp.core.Items, ConstProxyObj)
        assert isinstance(HABApp.core.EventBus, ConstProxyObj)

        # prevent we're calling from asyncio - this works because we don't use threads
        self.ctx = async_context.set("Rule Runner")

        ir = ItemRegistry()
        eb = EventBus()
        self.restore = setup_internals(ir, eb, final=False)

        # Scheduler
        self.monkeypatch.setattr(prod_sun_module, "OBSERVER", Observer(52.51870523376821, 13.376072914752532, 10))

        # Overwrite
        self.monkeypatch.setattr(HABApp.core, "EventBus", eb)
        self.monkeypatch.setattr(HABApp.core, "Items", ir)

        # Patch the hook so we can instantiate the rules
        hook = HABAppRuleHook(self.loaded_rules.append, suggest_rule_name, DummyRuntime(), None)
        self.monkeypatch.setattr(rule_module, "_get_rule_hook", lambda: hook)

        # patch worker with a synchronous worker
        self.monkeypatch.setattr(wrapped_thread, "POOL", self)
        self.monkeypatch.setattr(wrapper, "SYNC_CLS", WrappedThreadFunction, raising=False)

        # raise exceptions during error formatting
        self.monkeypatch.setattr(HABApp.core.lib.exceptions.format, "fallback_format", raising_fallback_format)

        # patch scheduler, so we run synchronous
        self.monkeypatch.setattr(job_builder_module, "AsyncHABAppScheduler", SyncScheduler)

    def tear_down(self) -> None:
        for rule in self.loaded_rules:
            rule._habapp_ctx.unload_rule()
        self.loaded_rules.clear()

        # restore patched
        self.monkeypatch.undo()

        # restore async context
        async_context.reset(self.ctx)
        self.ctx = None

        for r in self.restore:
            r.restore()

    def process_events(self) -> None:
        for s in SyncScheduler.ALL:
            for job in s.jobs:
                job.executor.execute()

    def __enter__(self) -> None:
        self.set_up()

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> bool:
        self.tear_down()
        # do not supress exception
        return False



"""Config models for heating rules."""

import datetime

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class KnxHeatingItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for KNX heating abstraction rule."""

    virtual_temperature: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="temperature item, which is used in OpenHAB to set the target temperature")
    actor_feedback_temperature: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="temperature item, which holds the current target temperature set by the heating actor")
    temperature_offset: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item for setting the offset temperature")


class KnxHeatingConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for KNX heating abstraction rule."""

    items: KnxHeatingItems = pydantic.Field(..., description="items for heating rule")
    parameter: None = None


class HeatingActiveItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for active heating rule."""

    control_values: list[HABApp.openhab.items.NumberItem] = pydantic.Field(..., description="list of control value items")
    output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item, which is ON when at least one control value is above the threshold")


class HeatingActiveParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameters for active heating rule."""

    threshold: float = pydantic.Field(0, description="control value threshold")
    extended_active_time: datetime.timedelta = pydantic.Field(datetime.timedelta(days=1), description="extended time to keep the output item ON, after last control value change below the threshold")


class HeatingActiveConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for active heating rule."""

    items: HeatingActiveItems = pydantic.Field(..., description="items for active heating rule")
    parameter: HeatingActiveParameter = pydantic.Field(HeatingActiveParameter(), description="parameters for active heating rule")



"""Common helper functions for all rules."""

import logging
import time

import HABApp
import HABApp.openhab.connection.handler.func_sync
import HABApp.openhab.items

import habapp_rules.core.exceptions

LOGGER = logging.getLogger(__name__)


def create_additional_item(name: str, item_type: str, label: str | None = None, groups: list[str] | None = None) -> HABApp.openhab.items.OpenhabItem:
    """Create additional item if it does not already exist.

    Args:
        name: Name of item
        item_type: Type of item (e.g. String)
        label: Label of the item
        groups: in which groups is the item

    Returns:
        returns the created item

    Raises:
        habapp_rules.core.exceptions.HabAppRulesError: if item could not be created
    """
    if not name.startswith("H_"):
        LOGGER.warning(f"Item '{name}' does not start with 'H_'. All automatically created items must start with 'H_'. habapp_rules will add 'H_' automatically.")
        name = f"H_{name}"

    if not HABApp.openhab.interface_sync.item_exists(name):
        if not label:
            label = f"{name.removeprefix('H_').replace('_', ' ')}"
        if not HABApp.openhab.interface_sync.create_item(item_type=item_type, name=name, label=label, groups=groups):
            msg = f"Could not create item '{name}'"
            raise habapp_rules.core.exceptions.HabAppRulesError(msg)
        time.sleep(0.05)
    return HABApp.openhab.items.OpenhabItem.get_item(name)


def send_if_different(item: str | HABApp.openhab.items.OpenhabItem, value: str | float) -> None:
    """Send command if the target value is different to the current value.

    Args:
        item: name of OpenHab item
        value: value to write to OpenHAB item
    """
    if isinstance(item, str):
        item = HABApp.openhab.items.OpenhabItem.get_item(item)

    if item.value != value:
        item.oh_send_command(value)


def filter_updated_items(input_items: list[HABApp.openhab.items.OpenhabItem], filter_time: int | None = None) -> list[HABApp.openhab.items.OpenhabItem]:
    """Get input items depending on their last update time and _ignore_old_values_time.

    Args:
        input_items: all items which should be checked for last update time
        filter_time: threshold for last update time

    Returns:
        full list if _ignore_old_values is not set, otherwise all items where updated in time.
    """
    if filter_time is None:
        return input_items

    filtered_items = [item for item in input_items if item.last_update.newer_than(filter_time)]

    if len(input_items) != len(filtered_items):
        ignored_item_names = [item.name for item in input_items if item.last_update.older_than(filter_time)]
        LOGGER.warning(f"The following items where not updated during the last {filter_time}s and will be ignored: {ignored_item_names}")

    return filtered_items



"""current switch rules."""

from typing import TYPE_CHECKING

import HABApp

import habapp_rules.core.helper
import habapp_rules.sensors.config.current_switch

if TYPE_CHECKING:
    from eascheduler.jobs.job_countdown import CountdownJob  # pragma: no cover


class CurrentSwitch(HABApp.Rule):
    """Rules class to manage basic light states.

    # Items:
    Number    Current              "Current"
    Switch    Something_is_ON      "Something is ON"

    # Config:
    config = habapp_rules.sensors.config.current_switch.CurrentSwitchConfig(
            items = habapp_rules.actors.config.light.CurrentSwitchItems(
                    current="Current",
                    switch="Something_is_ON"
            )
    )

    # Rule init:
    habapp_rules.actors.power.CurrentSwitch(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.current_switch.CurrentSwitchConfig) -> None:
        """Init current switch rule.

        Args:
            config: config for current switch rule
        """
        HABApp.Rule.__init__(self)
        self._config = config
        self._extended_countdown: CountdownJob | None = (
            self.run.countdown(self._config.parameter.extended_time, habapp_rules.core.helper.send_if_different, item=self._config.items.switch, value="OFF") if self._config.parameter.extended_time else None
        )

        self._check_current_and_set_switch(self._config.items.current.value)
        self._config.items.current.listen_event(self._cb_current_changed, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _check_current_and_set_switch(self, current: float | None) -> None:
        """Check if current is above the threshold and set switch.

        Args:
            current: current value which should be checked
        """
        if current is None:
            return

        current_above_threshold = current > self._config.parameter.threshold

        if self._config.parameter.extended_time:
            if current_above_threshold:
                self._extended_countdown.stop()
                habapp_rules.core.helper.send_if_different(self._config.items.switch, "ON")

            elif not current_above_threshold and self._config.items.switch.is_on():
                # start or reset the countdown
                self._extended_countdown.reset()

        else:
            # extended time is not active
            habapp_rules.core.helper.send_if_different(self._config.items.switch, "ON" if current_above_threshold else "OFF")

    def _cb_current_changed(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback, which is called if the current value changed.

        Args:
            event: event, which triggered this callback
        """
        self._check_current_and_set_switch(event.value)



"""Tests for sun sensors."""

import collections
import unittest.mock

import HABApp

import habapp_rules.sensors.config.sun
import habapp_rules.sensors.sun
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
from habapp_rules.system import PresenceState


class TestSensorTemperatureDifference(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing sun sensor 'temp_diff' rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Temperature_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_Temperature", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Threshold_Temperature", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "H_Temperature_diff_for_Unittest_Output_Temperature", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "H_Temperature_diff_for_Unittest_Output_Temperature_filtered", None)

        config = habapp_rules.sensors.config.sun.TemperatureDifferenceConfig(
            items=habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=["Unittest_Temperature_1", "Unittest_Temperature_2"], output="Unittest_Output_Temperature", threshold="Unittest_Threshold_Temperature")
        )

        with unittest.mock.patch("HABApp.openhab.interface_sync.item_exists", return_value=True), unittest.mock.patch("habapp_rules.common.filter.ExponentialFilter"):
            self._sensor = habapp_rules.sensors.sun.SensorTemperatureDifference(config)

    def test_init(self) -> None:
        """Test __init__."""
        self.assertEqual(float("inf"), self._sensor._hysteresis_switch._threshold)
        self.assertEqual("H_Temperature_diff_for_Unittest_Output_Temperature", self._sensor._item_temp_diff.name)

    def test_init_with_fixed_threshold(self) -> None:
        """Test __init__ with fixed threshold value."""
        config = habapp_rules.sensors.config.sun.TemperatureDifferenceConfig(
            items=habapp_rules.sensors.config.sun.TemperatureDifferenceItems(temperatures=["Unittest_Temperature_1", "Unittest_Temperature_2"], output="Unittest_Output_Temperature"),
            parameter=habapp_rules.sensors.config.sun.TemperatureDifferenceParameter(threshold=42),
        )

        with unittest.mock.patch("HABApp.openhab.interface_sync.item_exists", return_value=True), unittest.mock.patch("habapp_rules.common.filter.ExponentialFilter"):
            sensor = habapp_rules.sensors.sun.SensorTemperatureDifference(config)
        self.assertEqual(42, sensor._hysteresis_switch._threshold)

    def test_cb_threshold(self) -> None:
        """Test _cb_threshold."""
        tests.helper.oh_item.item_state_change_event("Unittest_Threshold_Temperature", 20)
        self.assertEqual(20, self._sensor._hysteresis_switch._threshold)

    def test_temp_diff(self) -> None:
        """Test if temperature difference is calculated correctly."""
        temp_diff_item = HABApp.openhab.items.OpenhabItem.get_item("H_Temperature_diff_for_Unittest_Output_Temperature")
        self.assertEqual(None, temp_diff_item.value)

        # update temperature 1
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_1", 20)
        self.assertEqual(None, temp_diff_item.value)

        # update temperature 2
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_2", 21)
        self.assertEqual(1, temp_diff_item.value)

        # update temperature 2
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_2", 18)
        self.assertEqual(2, temp_diff_item.value)

        # update temperature 1
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_1", -20)
        self.assertEqual(38, temp_diff_item.value)

        # update temperature 2
        tests.helper.oh_item.item_state_change_event("Unittest_Temperature_2", -25)
        self.assertEqual(5, temp_diff_item.value)

    def test_threshold_behavior(self) -> None:
        """Test overall behavior."""
        output_item = HABApp.openhab.items.OpenhabItem.get_item("Unittest_Output_Temperature")
        self.assertEqual(None, output_item.value)

        # set threshold to 10
        self._sensor._hysteresis_switch.set_threshold_on(10)

        # update temp_diff to 10
        tests.helper.oh_item.item_state_change_event("H_Temperature_diff_for_Unittest_Output_Temperature_filtered", 10)
        self.assertEqual("ON", output_item.value)

        # update temp_diff to 9.9
        tests.helper.oh_item.item_state_change_event("H_Temperature_diff_for_Unittest_Output_Temperature_filtered", 9.9)
        self.assertEqual("OFF", output_item.value)

        # update temp_diff to 8
        tests.helper.oh_item.item_state_change_event("H_Temperature_diff_for_Unittest_Output_Temperature_filtered", 8)
        self.assertEqual("OFF", output_item.value)


class TestSensorBrightness(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing sun sensor 'brightness' rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Brightness", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_Brightness", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Threshold_Brightness", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "H_Unittest_Brightness_filtered", None)

        config = habapp_rules.sensors.config.sun.BrightnessConfig(items=habapp_rules.sensors.config.sun.BrightnessItems(brightness="Unittest_Brightness", output="Unittest_Output_Brightness", threshold="Unittest_Threshold_Brightness"))

        with unittest.mock.patch("HABApp.openhab.interface_sync.item_exists", return_value=True), unittest.mock.patch("habapp_rules.common.filter.ExponentialFilter"):
            self._sensor = habapp_rules.sensors.sun.SensorBrightness(config)

    def test_init(self) -> None:
        """Test __init__."""
        self.assertEqual(float("inf"), self._sensor._hysteresis_switch._threshold)

    def test_init_with_fixed_threshold(self) -> None:
        """Test __init__ with fixed threshold value."""
        config = habapp_rules.sensors.config.sun.BrightnessConfig(
            items=habapp_rules.sensors.config.sun.BrightnessItems(
                brightness="Unittest_Brightness",
                output="Unittest_Output_Brightness",
            ),
            parameter=habapp_rules.sensors.config.sun.BrightnessParameter(threshold=42),
        )

        with unittest.mock.patch("HABApp.openhab.interface_sync.item_exists", return_value=True), unittest.mock.patch("habapp_rules.common.filter.ExponentialFilter"):
            sensor = habapp_rules.sensors.sun.SensorBrightness(config)
        self.assertEqual(42, sensor._hysteresis_switch._threshold)

    def test_cb_threshold(self) -> None:
        """Test _cb_threshold."""
        tests.helper.oh_item.item_state_change_event("Unittest_Threshold_Brightness", 42000)
        self.assertEqual(42000, self._sensor._hysteresis_switch._threshold)

    def test_threshold_behavior(self) -> None:
        """Test overall behavior."""
        output_item = HABApp.openhab.items.OpenhabItem.get_item("Unittest_Output_Brightness")
        self.assertEqual(None, output_item.value)

        # set threshold to 1000
        self._sensor._hysteresis_switch.set_threshold_on(1000)

        # update temp_diff to 1000
        tests.helper.oh_item.item_state_change_event("H_Unittest_Brightness_filtered", 1000)
        self.assertEqual("ON", output_item.value)

        # update temp_diff to 999
        tests.helper.oh_item.item_state_change_event("H_Unittest_Brightness_filtered", 999)
        self.assertEqual("OFF", output_item.value)

        # update temp_diff to 800
        tests.helper.oh_item.item_state_change_event("H_Unittest_Brightness_filtered", 800)
        self.assertEqual("OFF", output_item.value)


class TestSunPositionFilter(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing the sun position filter."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Input_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_1", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Input_2", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_2", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Azimuth", 1000)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Elevation", 1000)

        self.position_window_1 = habapp_rules.sensors.config.sun.SunPositionWindow(10, 80, 2, 20)
        self.position_window_2 = habapp_rules.sensors.config.sun.SunPositionWindow(100, 120)

        config_1 = habapp_rules.sensors.config.sun.SunPositionConfig(
            items=habapp_rules.sensors.config.sun.SunPositionItems(
                azimuth="Unittest_Azimuth",
                elevation="Unittest_Elevation",
                input="Unittest_Input_1",
                output="Unittest_Output_1",
            ),
            parameter=habapp_rules.sensors.config.sun.SunPositionParameter(sun_position_window=self.position_window_1),
        )
        config_2 = habapp_rules.sensors.config.sun.SunPositionConfig(
            items=habapp_rules.sensors.config.sun.SunPositionItems(
                azimuth="Unittest_Azimuth",
                elevation="Unittest_Elevation",
                input="Unittest_Input_2",
                output="Unittest_Output_2",
            ),
            parameter=habapp_rules.sensors.config.sun.SunPositionParameter(sun_position_window=[self.position_window_1, self.position_window_2]),
        )

        self._filter_1 = habapp_rules.sensors.sun.SunPositionFilter(config_1)
        self._filter_2 = habapp_rules.sensors.sun.SunPositionFilter(config_2)

    def test_init(self) -> None:
        """Test __init__."""
        self.assertEqual([self.position_window_1], self._filter_1._config.parameter.sun_position_windows)
        self.assertEqual([self.position_window_1, self.position_window_2], self._filter_2._config.parameter.sun_position_windows)

    def test_filter(self) -> None:
        """Test if filter is working correctly."""
        TestCase = collections.namedtuple("TestCase", "azimuth, elevation, input, output_1, output_2")

        test_cases = [
            TestCase(0, 0, "OFF", "OFF", "OFF"),
            TestCase(0, 10, "OFF", "OFF", "OFF"),
            TestCase(50, 0, "OFF", "OFF", "OFF"),
            TestCase(50, 10, "OFF", "OFF", "OFF"),
            TestCase(0, 0, "ON", "OFF", "OFF"),
            TestCase(0, 10, "ON", "OFF", "OFF"),
            TestCase(50, 0, "ON", "OFF", "OFF"),
            TestCase(50, 10, "ON", "ON", "ON"),
            TestCase(0, 0, "OFF", "OFF", "OFF"),
            TestCase(0, 10, "OFF", "OFF", "OFF"),
            TestCase(110, 0, "OFF", "OFF", "OFF"),
            TestCase(110, 10, "OFF", "OFF", "OFF"),
            TestCase(0, 0, "ON", "OFF", "OFF"),
            TestCase(0, 10, "ON", "OFF", "OFF"),
            TestCase(110, 0, "ON", "OFF", "ON"),
            TestCase(110, 10, "ON", "OFF", "ON"),
            TestCase(50, None, "OFF", "OFF", "OFF"),
            TestCase(None, 10, "OFF", "OFF", "OFF"),
            TestCase(None, None, "OFF", "OFF", "OFF"),
            TestCase(50, None, "ON", "ON", "ON"),
            TestCase(None, 10, "ON", "ON", "ON"),
            TestCase(None, None, "ON", "ON", "ON"),
        ]

        item_output_1 = HABApp.openhab.items.OpenhabItem.get_item("Unittest_Output_1")
        item_output_2 = HABApp.openhab.items.OpenhabItem.get_item("Unittest_Output_2")

        with unittest.mock.patch.object(self._filter_1, "_instance_logger") as log_1_mock, unittest.mock.patch.object(self._filter_2, "_instance_logger") as log_2_mock:
            for test_case in test_cases:
                log_1_mock.reset_mock()
                log_2_mock.reset_mock()

                tests.helper.oh_item.set_state("Unittest_Input_1", test_case.input)
                tests.helper.oh_item.set_state("Unittest_Input_2", test_case.input)

                tests.helper.oh_item.item_state_change_event("Unittest_Elevation", test_case.elevation)
                tests.helper.oh_item.item_state_change_event("Unittest_Azimuth", test_case.azimuth)

                self.assertEqual(test_case.output_1, item_output_1.value)
                self.assertEqual(test_case.output_2, item_output_2.value)

                if test_case.azimuth is None or test_case.elevation is None:
                    log_1_mock.warning.assert_called_once()
                    log_2_mock.warning.assert_called_once()
                else:
                    log_1_mock.warning.assert_not_called()
                    log_2_mock.warning.assert_not_called()


class TestWinterFilter(tests.helper.test_case_base.TestCaseBase):
    """Tests cases WinterFilter rule."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Sun", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Winter", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_1", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Output_2", None)

        config_full = habapp_rules.sensors.config.sun.WinterFilterConfig(
            items=habapp_rules.sensors.config.sun.WinterFilterItems(
                sun="Unittest_Sun",
                heating_active="Unittest_Winter",
                presence_state="Unittest_Presence_state",
                output="Unittest_Output_1",
            )
        )

        config_only_heating = habapp_rules.sensors.config.sun.WinterFilterConfig(
            items=habapp_rules.sensors.config.sun.WinterFilterItems(
                sun="Unittest_Sun",
                heating_active="Unittest_Winter",
                output="Unittest_Output_2",
            )
        )

        self._rule_full = habapp_rules.sensors.sun.WinterFilter(config_full)
        self._rule_winter = habapp_rules.sensors.sun.WinterFilter(config_only_heating)

    def test_filter(self) -> None:
        """Test WinterFilter rule."""
        TestCase = collections.namedtuple("TestCase", "sun, heating_active, presence_state, out_full, out_winter")

        test_cases = [
            # sun off
            TestCase(sun="OFF", heating_active="OFF", presence_state=PresenceState.PRESENCE, out_full="OFF", out_winter="OFF"),
            TestCase(sun="OFF", heating_active="OFF", presence_state=PresenceState.ABSENCE, out_full="OFF", out_winter="OFF"),
            TestCase(sun="OFF", heating_active="ON", presence_state=PresenceState.PRESENCE, out_full="OFF", out_winter="OFF"),
            TestCase(sun="OFF", heating_active="ON", presence_state=PresenceState.ABSENCE, out_full="OFF", out_winter="OFF"),
            # sun on
            TestCase(sun="ON", heating_active="OFF", presence_state=PresenceState.PRESENCE, out_full="ON", out_winter="ON"),
            TestCase(sun="ON", heating_active="OFF", presence_state=PresenceState.ABSENCE, out_full="ON", out_winter="ON"),
            TestCase(sun="ON", heating_active="ON", presence_state=PresenceState.PRESENCE, out_full="ON", out_winter="OFF"),
            TestCase(sun="ON", heating_active="ON", presence_state=PresenceState.ABSENCE, out_full="OFF", out_winter="OFF"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.item_state_change_event("Unittest_Sun", test_case.sun)
                tests.helper.oh_item.item_state_change_event("Unittest_Winter", test_case.heating_active)
                tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", test_case.presence_state.value)

                tests.helper.oh_item.assert_value("Unittest_Output_1", test_case.out_full)
                tests.helper.oh_item.assert_value("Unittest_Output_2", test_case.out_winter)



"""Config models for HCL color rules."""

import operator

import HABApp
import pydantic
import typing_extensions

import habapp_rules.core.pydantic_base


class HclTimeItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for HCL color which depends on time."""

    color: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="HCL color which will be set by the HCL rule")
    manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="switch item to disable all automatic functions")
    sleep_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleep state item")
    focus: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="focus state item")
    switch_on: HABApp.openhab.items.SwitchItem | HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="switch item which triggers a color update if switched on")
    state: HABApp.openhab.items.StringItem | None = pydantic.Field(..., description="state item for storing the current state")


class HclElevationItems(HclTimeItems):
    """Items for HCL color which depends on sun elevation."""

    elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun elevation")


class HclElevationParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for HCL color."""

    color_map: list[tuple[int, int]] = pydantic.Field([(-15, 3900), (0, 4500), (5, 5500), (15, 6500)], description="Color mapping. The first value is the sun elevation, the second is the HCL color")
    hand_timeout: int = pydantic.Field(18_000, description="hand timeout. After this time the HCL light rule will fall back to auto mode", gt=0)  # 5 hours
    sleep_color: float = pydantic.Field(2500, description="color if sleeping is active", gt=0)
    post_sleep_timeout: int = pydantic.Field(1, description="time after sleeping was active where the sleeping color will be set", gt=0)
    focus_color: float = pydantic.Field(6000, description="color if focus is active", gt=0)

    @pydantic.model_validator(mode="after")
    def validate_model(self) -> typing_extensions.Self:
        """Sort color map.

        Returns:
             model with sorted color map
        """
        self.color_map = sorted(self.color_map, key=operator.itemgetter(0))
        return self


_DEFAULT_TIME_MAP = [(0, 2200), (4, 2200), (5, 3200), (6, 3940), (8, 5000), (12, 7000), (19, 7000), (21, 5450), (22, 4000), (23, 2600)]


class HclTimeParameter(HclElevationParameter):
    """Parameter for HCL color which depends on time."""

    color_map: list[tuple[int, int]] = pydantic.Field(_DEFAULT_TIME_MAP, description="Color mapping. The first value is the hour, the second is the HCL color")
    shift_weekend_holiday: bool = pydantic.Field(default=False, description="If this is active the color will shift on weekends and holidays for one hour")


class HclElevationConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for HCL color which depends on sun elevation."""

    items: HclElevationItems = pydantic.Field(..., description="items for HCL color which depends on sun elevation")
    parameter: HclElevationParameter = pydantic.Field(HclElevationParameter(), description="parameter for HCL color which depends on sun elevation")


class HclTimeConfig(HclElevationConfig):
    """Config for HCL color which depends on time."""

    items: HclTimeItems = pydantic.Field(..., description="items for HCL color which depends on time")
    parameter: HclTimeParameter = pydantic.Field(HclTimeParameter(), description="parameter for HCL color which depends on time")



"""Config models for watchdog rules."""

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class WatchdogItems(habapp_rules.core.pydantic_base.ItemBase):
    """Items for watchdog rule."""

    observed: HABApp.openhab.items.OpenhabItem = pydantic.Field(..., description="observed item")
    warning: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="warning item, which will be set to ON if the observed item was not updated in the expected time")


class WatchdogParameter(habapp_rules.core.pydantic_base.ParameterBase):
    """Parameter for watchdog rule."""

    timeout: int = pydantic.Field(3600, description="timeout in seconds")


class WatchdogConfig(habapp_rules.core.pydantic_base.ConfigBase):
    """Config for watchdog rule."""

    items: WatchdogItems = pydantic.Field(..., description="items for watchdog rule")
    parameter: WatchdogParameter = pydantic.Field(WatchdogParameter(), description="parameters for watchdog rule")



"""Exceptions for HabAppRules."""


class HabAppRulesError(Exception):
    """Exception which is raised by this package."""


class HabAppRulesConfigurationError(HabAppRulesError):
    """Exception which is raised if wrong configuration is given."""
